
# $Id: Summary.py,v 1.15 2006/05/21 22:06:57 bmoe Exp $

import sqlobject
from HTMLgen import *
from HTMLgen.HTMLgen import Href, Heading, BR, Div, Caption, TableLite, TR, TD, TH, Form, Input, Select, Paragraph, Table, Strong
from HTMLgen.HTMLcolors import *
from BOSS import BossJob, BossConfig, config
from DAG import DAG
from DAGImage import DAGImage

from mx.DateTime import gmtime, DateTimeDelta, TimeDelta, localtime
from mx.DateTime.Parser import DateFromString

from Onasys import OnasysDocument

import OnasysBarchart
import os
imageDir = os.path.expandvars("$IMAGE_URL")
OnasysBarchart.barfiles = {
    'blue'   : imageDir + '/bar-blue.gif',
    'red'    : imageDir + '/bar-red.gif',
    'green'  : imageDir + '/bar-green.gif',
    'orange' : imageDir + '/bar-orange.gif',
    'yellow' : imageDir + '/bar-yellow.gif',
    'purple' : imageDir + '/bar-purple.gif'
}


import string

statusColor = { 'success': 'green',
                'failure' : 'red',
                'running' : 'lightblue',
                'unscheduled' : 'grey',
                'other' : 'yellow' }

class TCell(object):
    def __init__(self, val, attrs=""):
        self.val = val
        self.cellattr = attrs
    def __str__(self):
        return str(self.val)

def bossJobsInRange(start, end=None, user=None):
    if not end:
        end = gmtime()
        #start <= t_sub < end
    if user:
        userQ = BossJob.q.s_usr == user
    else:
        userQ = 1
    try:
        r = BossJob.select(sqlobject.AND(
                start.ticks() < BossJob.q.t_sub,
                BossJob.q.t_sub <= end.ticks(), userQ), orderBy="-t_sub")
    except:
        r = []
    return r


def bossJobsOnDay(date, user=None):
    if type(date) == str:
        date = DateFromString(date)
    oneDay = DateTimeDelta(1,0,0,0)
    return bossJobsInRange(date, date+oneDay, user=user)

def linkBack(cgi, text, action, vars={}, pageName=None):
    try:
        if not pageName:
            pageName = cgi.name

        href = "%s%s?function=%s" % (cgi.baseCGIRef, pageName, action)
        v = []
        for key in vars:
            v.append('%s=%s' % (key, vars[key]))
        if v:
            href += '&%s' % string.join(v,'&')
    except:
        return text
    return Href(text=text, url=href)
    

def barchart(ndays, cgi=None, user=None):
    chart = OnasysBarchart.StackedBarChart()
    chart.colors = [ 'green', 'red', 'blue', 'yellow' ]
    chart.label_shade = GRAY2
    chart.value_shade = GRAY3
    chart.title = "Job Summary For the Last %d Days" % ndays
    chart.datalist = OnasysBarchart.DataList()
    chart.datalist.segment_names = ['Successful', 'Failed', 'Running','Unscheduled']
    data = chart.datalist

    day = DateFromString('today')
    oneDay = DateTimeDelta(1,0,0,0)
    for i in range(0,ndays):
        jobs = bossJobsOnDay(day, user=user)
        success = 0
        error = 0
        running = 0
        other = 0
        for job in jobs:
            if job.t_stat == 'T':
                if job.ret_code == "0":
                    success += 1
                else:
                    error += 1
            elif job.t_stat == 'E': error += 1
            elif job.t_stat == 'R': running += 1
            else: other += 1
        timeLabel = day.strftime("%b %d")
        if cgi:
            if user: userVar = "&user=%s" % user
            else: userVar = ""
            timeHref = cgi.baseCGIRef + "Summary?function=displayDay&day=%s%s"%(day.strftime("%Y-%m-%d"), userVar)
            timeLink = Href(text=timeLabel, url=timeHref)
        else:
            # No link.  Just give label.
            timeLink = timeLabel
        data.load_tuple((timeLink, success, error, running, other))
        day = day - oneDay
    return chart


#------------------------------------------------------------------
# Monitor Stuff.

import pickle
import os
import time

def findInfoFiles(path):
    infoFiles = []
    n=0
    for root, dirs, files in os.walk(path):
        n += 1
        for file in files:
            if file[-5:] == '.info':
                #remember this file
                infoFiles.append(os.path.join(root,file))
                # don't descend this directory tree any further.
                dirs[:] = []
    return infoFiles

def unpackInfoFile(fname):
    try:
        f = open(fname,'r')
    except:
	raise
#        return None

    d = pickle.load(f)
    t = os.stat(fname).st_mtime
    d['access'] = t

    splitName = fname.split('.')
    analysisName = splitName[1]

    return (analysisName, d)

def monitorTable(infoDir):
    table = TableLite(border="1")
    row = TR()
    for heading in ["Analysis Name", "*", "Last Run<br>(seconds ago)", "Contact"]:
        row.append(TH(heading, html_escape="NO"))
    table.append(row)

    for f in findInfoFiles(infoDir):
        row = TR()
        d = unpackInfoFile(f)
        name = d[0]
        contact = d[1]['e-mail']
	contactLink = '<a href="mailto:%s">%s</a>' % (contact, contact)
        lastRun = time.time() - d[1]['access']
        updateInterval = d[1]['update_interval']
        if lastRun > 1.1 * updateInterval + 360:
    	    color = "red"
        elif lastRun > updateInterval + 5:
    	    color = "yellow"
        else:
    	    color = "green"

        row.append(TD(name))
        row.append(TD("&nbsp;", html_escape="NO", bgcolor=color))
        row.append(TD("%0.0f"%lastRun, align="right"))
        row.append(TD(contactLink, html_escape="NO"))
        
        table.append(row)
    return table


class Summary(OnasysDocument):
    def __init__(self, path=None, *args, **kw):
        self.maxdisplaynodes = 10
        OnasysDocument.__init__(self, *args, **kw)
        self.name = "Summary"
        self.thePath = path
        self.script = [checkBoxJavascript()]
        #self.title = "Onasys Monitor"
        self.valid = {'_no_function': self.default,
                      'display': self.display,
                      'displayDay': self.displayDay,
                      'displayDAG': self.displayDAG,
                      'displayNode' : self.displayNode,
                      'daemonMon': self.daemonMon,
                     }
        nav = HTMLgen.NonBulletList()
        self.nav = nav
        nav.append(linkBack(self, "Today's &nbsp;Jobs", "displayDay"))
        nav.append(linkBack(self, "Recent&nbsp;Jobs", "display"))
        nav.append(linkBack(self, "Daemon&nbsp;Monitor", "daemonMon"))


    def default(self, cgi):
#        import popen2
#        oo,ii = popen2.popen2('/bin/cat /etc/hostname.skge0 /etc/hostname.skge1')
#        self.append(oo.read())
#        oo.close()
#        ii.close()
        self.append("Welcome")

    def display(self, cgi):
        #form = Form(submit=Input(name='function', value='display', type='hidden'))
        form = Form(method="GET", submit=Input(name='submit', value='Go', type='submit'))
        try:
            days = int(self.form['days'])
        except:
            days = 0

        try:
            user = self.form['user']
            if user == "ANY":
                user = None
        except:
            user = "ANY"

        users = self._getUsers()
        #users = self.users
        #['kipp', 'dbrown', 'lindy', 'siemens', 'lgoggin' ]

        pullDays = Select(data=[0,2,5,10,15,20,25,30], name='days', size=1, selected=[days])
        pullUser = Select(data=['ANY']+users, name='user', size=1, selected=[user])

        #t = TableLite()
        #r = TR()
        form.append("Display summary for the last N Days:")
        form.append(pullDays)
        form.append("For user:")
        form.append(pullUser)

        form.append(Input(name='function', value='display', type='hidden'))
        self.append(form)
        try:
            d = int(self.form['days'])
        except:
            d = 0
        if d:
            try:
                self.append(barchart(d, self, user=user))
            except Exception, e:
                self.append(str(e))

    def displayDay(self, cgi):
        try:
            day = self.form['day']
            dayLabel = DateFromString(day).strftime("%a %b %d, %Y")
        except:
            day = "Today"
            dayLabel = "Today &mdash; " + DateFromString('today').strftime("%a %b %d, %Y")

        user = None
        userLabel = ""
        try:
            user = self.form['user']
            if user == "ALL":
                user = None
            else:
                userLabel = "User: %s" % user
        except: pass

	try: 
            jobs = bossJobsOnDay(day, user=user)
            h = Heading(1,"Job Summary for: %s %s" % (dayLabel, userLabel), html_escape="NO")
            self.append(h)
            self.append(checkBoxes())
            self.append(BR())
            self.append(bossJobTable(jobs, cgi=self, dateFormat="%H:%M:%S"))
	except Exception, e:
	    self.append(Paragraph("Cannot get information: %s" % str(e)))

    def displayDAG(self, cgi):
        jobid = self.form['jobid']
        self.append(Heading(1,"Summary of DAG for Job #%s" % jobid))

        # Get the DAG itself.
        dag = DAG.get(jobid=jobid)
        if not dag:
            self.append(Strong("No DAG for this job"))
            return

	try:
		dagtable = dagTable(dag, title="", cgi=self)
	except Exception, e:
		self.append(String("Cannot make dag table. (%s)"%str(e)))
		return

        imagemap = ''
        imagemapname = "MAP%s" % jobid
        ref = '%s%s?function=displayNode&jobid=%s&node=\\N' % (cgi.baseCGIRef, self.name, jobid)

	try:
		imagemap = DAGImage(jobid).imageMap(mapname=imagemapname, baseref=ref)

		dagimage = Image(cgi.baseCGIRef+"DAGImage/%s" % jobid,
				 'DAG Image',
				 absolute=self.baseCGIRef,
				 border=0,
				 usemap=imagemapname)
	except Exception, e:
		dagimage = Strong("No DAG image available")
		imagemap = Paragraph(str(e))


        self.append(dagtable)
        if len(dag.nodes()) > self.maxdisplaynodes:
            self.append(Strong("Too many nodes to display"))
        else:
            self.append(dagimage)
            self.append(imagemap)

    def displayNode(self, cgi):
        jobid = cgi.form['jobid']
        nodeName = cgi.form['node']

        dag = DAG.get(jobid=jobid)

        node = dag.node(nodeName)

        if not node:
            self.append(Strong("ERROR:"))
            self.append("Cannot find node '%s' for job '%s'" %(jobid, nodeName))
            return

        t = TableLite(width="")
        self.append(Heading(1,"Node %s / Job # %s" % (nodeName, jobid)))

        row = TR()
        row.append(TD(Strong("Job ID:"), align="right"))
        row.append(TD( linkBack(self, node.jobid, 'displayDAG', vars={'jobid':jobid})))
        t.append(row)

        columns = [
                    ('name', "Name"),
                    ('t_sub', "Submit Time"),
                    ('t_start', "Start Time"),
                    ('t_stop', "Stop Time"),
                    ('ret_code', "Return Code"),
                    ('subfile', "Submit File"),
                    ('stdout', "Std Output File"),
                    ('stderr', "Std Error File"),
                  ]
        for key in columns:
            row = TR()
            row.append(TD(Strong(key[1]+":"), align='right'))
            row.append(TD(getattr(node, key[0])))
            t.append(row)
        self.append(t)

        self.append(BR())
        if len(node.parents):
            p = Paragraph()
            p.append(Strong("Parent Nodes"))
            p.append(BR())
            for parent in node.parents:
                vars = { 'jobid' : jobid, 'node' : parent.name }
                p.append(linkBack(self, parent.name, 'displayNode', vars=vars))
                p.append(Strong("&nbsp; ", html_escape="NO"))
            self.append(p)

        if len(node.children):
            p = Paragraph()
            p.append(Strong("Child Nodes"))
            p.append(BR())
            for child in node.children:
                vars = { 'jobid' : jobid, 'node' : child.name }
                p.append(linkBack(self, child.name, 'displayNode', vars=vars))
                p.append(Strong("&nbsp; ", html_escape="NO"))
            self.append(p)

    def daemonMon(self, cgi):
	t = monitorTable(self.monitorInfoDir)
        p = Paragraph(t, html_escape="NO")
	self.append(p)
	return

    def _getUsers(self):
        # This is a pain in the behind.
        # It is also a terrible hack.  SQLObject doesn't seem
        # to want to let you do this.  Sadness!
        # It is also dangerous in that it is MySQL specific.
        # XXX Should probably test for that MySQL-ness and
        # substitute the contents of self.users or something
        # if it's not MySQL.
        conn = BossJob._connection.getConnection()
        cursor = conn.cursor()
        cursor.execute('SELECT DISTINCT S_USR FROM JOB')
        result = [ x for (x,) in cursor.fetchall() ]
        cursor.close()
        del cursor
        conn.close()
        del conn
        return result

def bossJobTable(jobs, title="", cgi=None, dateFormat=None, tableName="table"):
    t = TableLite(cellspacing="1", cellpadding="2", border="1", id=tableName)
    badNodeTables = Div()
    t.append(Caption(title))
    headings = ["ID",
                 "SID",
                 "*",
                 "Host",
                 "User",
                 "Path",
                 "Submit Time",
                 "Start Time",
                 "Stop Time",
                 "Exec",
                 "Status",
                 "Return Code",
                 ]
    row = TR()
    for heading in headings:
        row.append(TH(heading))
    t.append(row)

    for job in jobs:

#        # If we're a failed job, collect the names of failed dag nodes.
#        if (nodeColor(job) == 'red'):
#            # There ought to be a better way to find this than looking at COLOR
#            #badTable = TableLite(id="pop_%s"%job.id)
#            #badTable.__setattr__('class', 'hidden')
#            badTable = TableLite()
#            badTable.__setattr__('class', 'menu small')
#            badTable.append(TR(TH("Bad Nodes")))
#            dag = DAG.get(jobid=job.id)
#            for node in dag.nodes():
#                if nodeColor(node) == "red":
#                    badTable.append(TR(TD(node.name)))
#            d = Div(badTable, id="pop_%s"%job.id)
#            d.__setattr__('class', 'hidden')
#            badNodeTables.append(d)
#            statusElement = TD("&nbsp;", html_escape="NO", bgcolor=nodeColor(job), onmouseover="displayNode(%s)"%job.id, onmouseout="hideNode(%s)"%job.id, id=job.id)
#            # GRR.  HTMLgen doesn't understant mouseovers in table cells.
#            # -- ok.  now that i hacked it in, it does.
#            #    seems table was putting our <TD> string into another td. :P
#            #statusElement = '<td bgcolor="%s" onmouseover="displayNode(%s)" onmouseout="hideNode(%s)">&nbsp;</td>' % (nodeColor(job), job.id, job.id)
#        else:
#            statusElement = TD("&nbsp;", html_escape="NO", bgcolor=nodeColor(job))
        statusElement = TD("&nbsp;", html_escape="NO", bgcolor=nodeColor(job))

        link = linkBack(cgi, job.id, "displayDAG", vars={'jobid':job.id})
        statusCell = TCell("", nodeColor(job, strPat='bgcolor="%s"'))
        row = TR()
        # OK.  This is lame.
        # row.class=blah fails.  TR(class=blah) fails.
        # WHY is 'class' a KEYWORD for gack's sake?
        # THAT is why I'm idiotically using __setattr__...
        # I just can't remember what freaking python magic
        # __setattr__ is accessed with. :P
        row.__setattr__('class', nodeColor(job))
        row.append(TD(link))
        row.append(TD(_cellval(job.sid)))
        row.append(statusElement)
        row.append(TD(_cellval(job.s_host)))
        row.append(TD(_cellval(job.s_usr)))
        row.append(TD(_cellval(job.s_path)))
        row.append(TD(_cellval(str(bossDate(job.t_sub, format=dateFormat)))))
        row.append(TD(_cellval(str(bossDate(job.t_start, format=dateFormat)))))
        row.append(TD(_cellval(str(bossDate(job.t_stop, format=dateFormat)))))
        row.append(TD(_cellval(job.executable)))
        row.append(TD(_cellval(job.t_stat)))
        row.append(TD(_cellval(job.ret_code), html_escape="NO"))
        t.append(row)

    retval = Div()
    retval.append(t)
    retval.append(badNodeTables)
    return retval


def nodeColor(node, strPat="%s"):
    if node.t_stat == 'T' and node.ret_code == '0':
        color = strPat % statusColor['success']
    elif node.t_stat == 'T':
        color = strPat % statusColor['failure']
    elif node.t_stat == 'R':
        color = strPat % statusColor['running']
    elif not node.t_sub:
        color = strPat % statusColor['unscheduled']
    else:
        color = strPat % statusColor['other']
    return color

def _cellval(x):
    if not x:
        return "&nbsp;"
    else:
        return x

def bossDate(x, format=None):
    if x == 0:
        return 0
    if not format:
        return localtime(x)
    else:
        return localtime(x).strftime(format)

def dagTable(dag, title="", cgi=None):
    t = Table(title, width="")
    t.body=[]
    t.heading = [
                 "Name",
                 "*",
                 "Submit Time",
                 "Start Time",
                 "Stop Time",
                 "Status",
                 "Return Code"
                ]
    for node in dag.nodes():
        if cgi:
            vars = { 'jobid': dag.jobid, 'node' : node.name }
            name = linkBack(cgi, node.name, 'displayNode', vars=vars)
        else:
            name = _cellval(node.name)

        statusCell = TCell("", nodeColor(node, strPat='bgcolor="%s"'))

        line = [
                name,
                statusCell,
                _cellval(node.t_sub),
                _cellval(node.t_start),
                _cellval(node.t_stop),
                _cellval(node.t_stat),
                _cellval(node.ret_code),
               ]
        t.body.append(line)
    return t

def checkBoxJavascript():
    script = """
<SCRIPT LANGUAGE="javascript">
<!--
    function changeVisible(table)
    {
        for (i=0; i < document.getElementById(table).rows.length; i++)
        {
            row = document.getElementById(table).rows[i]
            box = document.getElementById(row.className)
            if (box && box.checked)
            {
                row.style.display = "";
            }
            else if (box)
            {
                row.style.display = "none";
            }
        }
    }
    function find_position(element) {
    // from webreference.com by way of mythtv
    // and post brutal hacks to promote browser incompatibility.
    // No parent, just return the coordinates
        isIE4 = 0;
        isIE6 = 0;
        if (! element.offsetParent)
            return [element.x, element.y];
    // Scan backwards through the parents
        for( var x = 0, y = 0; element.offsetParent; element = element.offsetParent ) {
        // If IE...
            if (isIE4 || isIE6) {
                // If element is not a table or body tag, append the cell border info
                if (element.tagName != 'TABLE' && element.tagName != 'BODY') {
                    x += element.clientLeft
                    x += element.clientTop
                }
            }
        // Gecko?
            else {
            // We need to take the table border into consideration
                if (element.tagName == 'TABLE') {
                    var border = parseInt(element.border);
                // No visible border, check for a frame attribute
                    if (isNaN(border)) {
                        var frame = element.getAttribute('frame');
                    // Found a frame attribute, but only add one pixel for it.
                        if (frame != null) {
                            x++;
                            y++;
                        }
                    }
                // Visible border, add it to the calculation, too
                    else if (border > 0) {
                        x += border;
                        y += border;
                    }
                }
            }
        // Don't forget the actual location of the element
            x += element.offsetLeft;
            y += element.offsetTop;
        }
        return [x, y];
    }
    function displayNode(n)
    {
        name = "pop_" + n;
        element = document.getElementById("pop_" + n);
        posElement = document.getElementById(n);
        pos = find_position(posElement);
        //alert("POS: " + pos);
        var x = parseInt(pos[0]);
        var y = parseInt(pos[1]); + element.offsetHeight; 
        //alert("X:" + x + "\\nY:" + y);
        element.style.left = x + 30;  // + element width, really?
        element.style.top = y;
        element.style.visibility = "visible";
    }
    function hideNode(n)
    {
        element = document.getElementById("pop_" + n);
        element.style.visibility = "hidden";
    }
-->
</SCRIPT>
<STYLE>
    .menu { background-color: #d6d6d6 }
    .small { font-size: 12px; }
    .hidden  {
        visibility : hidden;
        position : absolute;
        z-index : 99 ;
        top = 0px;
        left = 0px;
        float = center;
    }
</STYLE>
    """
    return script

def checkBoxes(tableName='table'):
    boxes = Div(html_escape="NO")
    boxes.append("Display: &nbsp;")
    for label, type in [('Succeeded', 'success'),
                        ('Failed', 'failure' ),
                        ('Running', 'running' ),
                        ('Unscheduled', 'unscheduled' ),
                        ('Other', 'other' )]:
        #box = Input(id=statusColor[type], type='checkbox', checked='checked', llabel=label, onClick="changeVisible('%s')"%tableName)
        # ARGH.  HTMLgen does not let us put an id on an input element.
        # That is just annoying.
        box = """<INPUT ID="%s" TYPE="checkbox" NAME="default" CHECKED onClick="changeVisible('%s')">%s""" % (statusColor[type], tableName, label)
        boxes.append(box)
        #A<input id="a" class="radio" onClick="goober()" checked="checked" type="checkbox"><br>
    return boxes
    return ""

    pass
